/*
   bitset.c -- PIKT piktc bitset routines

   $Header: /home/cvs/ucgsb/pikt/src/piktc/bitset.c,v 1.13 1999/10/30 17:46:53 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
*/

#include "../lib/common.h"
#include "piktc.h"

unsigned long *ghbs;
unsigned long *gdbs, *gubs;
unsigned long *hbsfull, *hbsempty;
unsigned long *dbsfull, *dbsempty;

/* empty a bitset (turn off all bits)
 */
void
empty (unsigned long *bs, size_t bslongs)
{
	register int i;

	for (i=0; i<bslongs; i++)
		*(bs+i) = 0UL;
	return;
}

/* fill a bitset (turn on all bits)
 */
void
fill (unsigned long *bs, size_t bslongs)
{
	register int i;

	for (i=0; i<bslongs; i++)
		*(bs+i) = ~0UL;
	return;
}

/* union of two bitsets;
   returns a calloc()'ed bs that should eventually be free()'ed
 */
unsigned long *
bsunion (unsigned long *bs1, unsigned long *bs2, size_t bslongs)
{
	register int i;
	unsigned long *bs;

	bs = (unsigned long *) mycalloc(bslongs, sizeof(unsigned long), "bsunion");
	for (i=0; i<bslongs; i++)
		*(bs+i) = *(bs1+i) | *(bs2+i);
	return bs;
}

/* intersection of two bitsets;
   returns a calloc()'ed bs that should eventually be free()'ed
 */
unsigned long *
bsintersection (unsigned long *bs1, unsigned long *bs2, size_t bslongs)
{
	register int i;
	unsigned long *bs;

	bs = (unsigned long *) mycalloc(bslongs, sizeof(unsigned long), "bsintersection");
	for (i=0; i<bslongs; i++)
		*(bs+i) = *(bs1+i) & *(bs2+i);
	return bs;
}

/* complement of a bitset;
   returns a calloc()'ed bs that should eventually be free()'ed
 */
unsigned long *
bscomplement (unsigned long *bs1, size_t bslongs)
{
	register int i;
	unsigned long *bs;

	bs = (unsigned long *) mycalloc(bslongs, sizeof(unsigned long), "bscomplement");
	for (i=0; i<bslongs; i++) {
		*(bs+i) = ~*(bs1+i);
	}
	return bs;
}

/* is item (the second bitset--a sequence of 0's and just a single 1)
   a member of a set (the first bitset--a sequence of 0's and 1's)
 */
int
bsmember (unsigned long *bs1, unsigned long *bs2, size_t bslongs)
{
	register int i;
	static unsigned long *bs;

	bs = bsintersection(bs1, bs2, bslongs);
	for (i=0; i<bslongs; i++)
		if (*(bs+i)) {
			free(bs);
			return TRUE;
		}
	free(bs);
	return FALSE;
}

/* is bs2 a subset of bs1?  if the intersection of bs2 with bs1's complement
   is not the empty set, return false, else return true
 */
int
bssubset (unsigned long *bs1, unsigned long *bs2, size_t bslongs)
{
	register int i;
	static unsigned long *bs;

	bs = bscomplement(bs1, bslongs);
	for (i=0; i<bslongs; i++) {
		if (*(bs+i) & *(bs2+i)) {
			free(bs);
			return FALSE;
		}
	}
	free(bs);
	return TRUE;
}

/* is even just a single bit turned on in the given bit set?
 * /
unsigned long
defined (unsigned long *bs, size_t bslongs)
{
	register int i;
	unsigned long d;

	d = 0UL;
	for (i=0; i<bslongs; i++)
		d |= *(bs+i);
	return d;
}
*/

/* copy one hostset to another
 */
void
bscopy (unsigned long *bs1, unsigned long *bs2, size_t bslongs)
{
	register int i;

	for (i=0; i<bslongs; i++)
		*(bs1+i) = *(bs2+i);
	return;
}

/* duplicate a bitset, returning a pointer to the duplicate;
   returns a calloc()'ed bs that should eventually be free()'ed
 */
unsigned long *
bsdup (unsigned long *bs2, size_t bslongs)
{
	register int i;
	unsigned long *bs;

	bs = (unsigned long *) mycalloc(bslongs, sizeof(unsigned long), "bsdup");
	for (i=0; i<bslongs; i++)
		*(bs+i) = *(bs2+i);
	return bs;
}

/* get the bs (bitset) for the specified ident
 */
unsigned long *
getbs (char *identname, int type)
{
	ident *idp;

	switch (type) {
		case OPS:
		case HST:
		case GRP:
		case ALS:
			idp = lookupident(identname, SYSCFG);
			if (idp)
				return idp->bs;
			else {
				fprintf(stderr,"%s, at or near line %d: host, group, or alias %s unknown\n", cfgfilecurr, lineno, identname);
				fflush(stderr);
				remove(piktclock);
				exit(1);
				/* return NULL; */
			}
			break;
		case DEF:
			idp = lookupident(identname, DEFCFG);
			if (idp)
				return idp->bs;
			else {
				fprintf(stderr,"%s, at or near line %d: define %s unknown\n", cfgfilecurr, lineno, identname);
				fflush(stderr);
				remove(piktclock);
				exit(1);
				/* return NULL; */
			}
			break;
		default:
			return (unsigned long *) NULL;
			break;
	}
}

/* apply the bitset tag to the current config line, also its line number,
   cfg file, and length, and whether it should be macro expanded; mark it
   for processing by default
 */
void
tag (int lno, char *file, unsigned long *bs, size_t bslongs, int domcr, int doproc, char *line)
{
	static char _tag_[] = "tag";
	cfgln *nclp;

	nclp = (cfgln *) mymalloc(sizeof(cfgln), _tag_);
	nclp->next = (cfgln *) NULL;
	nclp->lineno = lno;
	if (clplast) {
		if (strcmp(clplast->file, file))
			nclp->file = mystrdup(file, _tag_);
		else
			nclp->file = clplast->file;
	} else
		nclp->file = mystrdup(file, _tag_);
	nclp->bs = bsdup(bs, bslongs);
	nclp->line = line;
	nclp->len = strlen(line);
	nclp->domcr = domcr;
	nclp->doproc = doproc;
	if (!configs[cfgcurr]) {
		configs[cfgcurr] = nclp;
	} else {
		clplast->next = nclp;
	}
	clplast = nclp;
}

/* insert the current ident into the global bitset
 */
ident *
insert (char *identname, int type)
{
	ident *idp;
	register int i;
	
	switch (type) {
		case HST:	/* or OPS, GRP, or ALS */
			idp = lookupident(identname, SYSCFG);
			if (idp) {
				for (i=0; i<hbslongs; i++)
					*(ghbs+i) |= *((idp->bs)+i);
			} else
				logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in insert(), ident %s not registered", identname);
			return idp;
			break;
		case DEF:

			idp = lookupident(identname, DEFCFG);
			if (idp) {
				for (i=0; i<dbslongs; i++)
					*(gdbs+i) |= *((idp->bs)+i);
			} else
				logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in insert(), define %s not registered", identname);
			return idp;
			break;
		default:
			return (ident *) NULL;
			break;
	}
}
